@@ -23,6 +23,8 @@
 from __future__ import absolute_import, unicode_literals
 
 from datetime import date
+
+import json
 from dateutil.relativedelta import relativedelta
 from dateutil.parser import parse
 import re
@@ -40,6 +42,32 @@ from inspire_query_parser.config import (DATE_LAST_MONTH_REGEX_PATTERN,
                                          DATE_YESTERDAY_REGEX_PATTERN)
 
 
+NAME_INITIAL_FOLLOWED_BY_FIRSTNAME_WITHOUT_SPACE = re.compile(r"(\.[a-z])", re.IGNORECASE)
+
+
+def retokenize_first_names(names):
+    """Handle corner cases where the intial and firstname has no space.
+
+    Example:
+        For queries ``J.David`` we be split into ``J`` and ``David``.
+    """
+    names_filtered = []
+    for name in names:
+        if not name:
+            continue
+
+        match = NAME_INITIAL_FOLLOWED_BY_FIRSTNAME_WITHOUT_SPACE.search(name)
+        if match:
+            names_filtered.extend(name.split('.'))
+        else:
+            names_filtered.append(name)
+    return names_filtered
+
+
+def is_initial_of_a_name(name_part):
+    return len(name_part) == 1 or u'.' in name_part
+
+
 def author_name_contains_fullnames(author_name):
     """Recognizes whether the name contains full name parts and not initials or only lastname.
 
@@ -47,14 +75,11 @@ def author_name_contains_fullnames(author_name):
           bool: True if name has only full name parts, e.g. 'Ellis John', False otherwise. So for example, False is
             returned for 'Ellis, J.' or 'Ellis'.
     """
-    def _is_initial(name_part):
-        return len(name_part) == 1 or u'.' in name_part
-
     parsed_name = ParsedName(author_name)
 
     if len(parsed_name) == 1:
         return False
-    elif any([_is_initial(name_part) for name_part in parsed_name]):
+    elif any([is_initial_of_a_name(name_part) for name_part in parsed_name]):
         return False
 
     return True
@@ -92,7 +117,7 @@ def generate_minimal_name_variations(author_name):
         lastnames = parsed_name.last.replace('-', ' ')
 
         non_lastnames = ' '.join(
-            parsed_name.first_list + parsed_name.middle_list + parsed_name.suffix_list
+            parsed_name.first_list + parsed_name.suffix_list
         )
         # Strip extra whitespace added if any of middle_list and suffix_list are empty.
         non_lastnames = non_lastnames.strip().replace('-', ' ')
@@ -374,8 +399,18 @@ def generate_match_query(field, value, with_operator_and):
     Notes:
         If value is of instance bool, then the shortened version of the match query is generated, at all times.
     """
+    parsed_value = None
+    try:
+        parsed_value = json.loads(value.lower())
+    except (ValueError, TypeError, AttributeError):
+        # Catch all possible exceptions
+        # we are not interested if they will appear
+        pass
+
     if isinstance(value, bool):
         return {'match': {field: value}}
+    elif isinstance(parsed_value, bool):
+        return {'match': {field: value.lower()}}
 
     if with_operator_and:
         return {
@@ -434,3 +469,21 @@ def wrap_queries_in_bool_clauses_if_more_than_one(queries,
             ('must' if use_must_clause else 'should'): queries
         }
     }
+
+
+def wrap_query_in_nested_if_field_is_nested(query, field, nested_fields):
+    """Helper for wrapping a query into a nested if the fields within the query are nested
+
+    Args:
+        query : The query to be wrapped.
+        field : The field that is being queried.
+        nested_fields : List of fields which are nested.
+    Returns:
+        (dict): The nested query
+    """
+    for element in nested_fields:
+        match_pattern = r'^{}.'.format(element)
+        if re.match(match_pattern, field):
+            return generate_nested_query(element, query)
+
+    return query
