@@ -27,7 +27,6 @@ visitor and converts it to an ElasticSearch query.
 
 from __future__ import absolute_import, unicode_literals
 
-from itertools import product
 import logging
 from pypeg2 import whitespace
 import re
@@ -36,7 +35,7 @@ from unicodedata import normalize
 
 from inspire_schemas.utils import convert_old_publication_info_to_new
 from inspire_utils.helpers import force_list
-from inspire_utils.name import normalize_name
+from inspire_utils.name import normalize_name, ParsedName
 
 from inspire_query_parser import ast
 from inspire_query_parser.config import (
@@ -47,12 +46,13 @@ from inspire_query_parser.utils.visitor_utils import (
     ES_RANGE_EQ_OPERATOR,
     _truncate_date_value_according_on_date_field,
     _truncate_wildcard_from_date,
-    author_name_contains_fullnames,
     generate_match_query,
-    generate_minimal_name_variations,
     generate_nested_query,
     update_date_value_in_operator_value_pairs_for_fieldname,
     wrap_queries_in_bool_clauses_if_more_than_one,
+    wrap_query_in_nested_if_field_is_nested,
+    is_initial_of_a_name,
+    retokenize_first_names
 )
 from inspire_query_parser.visitors.visitor_impl import Visitor
 
@@ -70,7 +70,6 @@ class ElasticSearchVisitor(Visitor):
     Notes:
         The ElasticSearch query follows the 2.4 version DSL specification.
     """
-
     # ##### Configuration #####
     # ## Journal queries ##
     JOURNAL_FIELDS_PREFIX = 'publication_info'
@@ -90,6 +89,9 @@ class ElasticSearchVisitor(Visitor):
     # TODO Inspire mappings aren't in their own repository. Currently using the `records-hep` mapping.
     KEYWORD_TO_ES_FIELDNAME = {
         'author': 'authors.full_name',
+        'author_first_name': 'authors.first_name',
+        'author_last_name': 'authors.last_name',
+        'author_first_name_initials': 'authors.first_name.initials',
         'author-count': 'author_count',
         'citedby': 'citedby',
         'collaboration': 'collaborations.value',
@@ -107,12 +109,13 @@ class ElasticSearchVisitor(Visitor):
         'journal': [
             JOURNAL_FIELDS_MAPPING.values()
         ],
-        'refersto': 'references.recid',
+        'refersto': 'references.record.$ref',
         'reportnumber': 'report_numbers.value.fuzzy',
         'subject': 'facet_inspire_categories',
         'title': 'titles.full_title',
         'type-code': 'document_type',
         'topcite': 'citation_count',
+        'affiliation': 'authors.affiliations.value',
     }
     """Mapping from keywords to ElasticSearch fields.
 
@@ -122,13 +125,22 @@ class ElasticSearchVisitor(Visitor):
     """
     TYPECODE_VALUE_TO_FIELD_AND_VALUE_PAIRS_MAPPING = {
         'b': ('document_type', 'book'),
+        'book': ('document_type', 'book'),
         'c': ('document_type', 'conference paper'),
+        'conferencepaper': ('document_type', 'conference paper'),
+        'citeable': ('citeable', True),
         'core': ('core', True),
         'i': ('publication_type', 'introductory'),
+        'introductory': ('publication_type', 'introductory'),
         'l': ('publication_type', 'lectures'),
+        'lectures': ('publication_type', 'lectures'),
         'p': ('refereed', True),
+        'published': ('refereed', True),
         'r': ('publication_type', 'review'),
+        'review': ('publication_type', 'review'),
         't': ('document_type', 'thesis'),
+        'thesis': ('document_type', 'thesis'),
+        'proceedings': ('document_type', 'proceedings'),
     }
     """Mapping from type-code query values to field and value pairs.
 
@@ -139,9 +151,16 @@ class ElasticSearchVisitor(Visitor):
     AUTHORS_NAME_VARIATIONS_FIELD = 'authors.name_variations'
     AUTHORS_BAI_FIELD = 'authors.ids.value'
     BAI_REGEX = re.compile(r'^((\w|-|\')+\.)+\d+$', re.UNICODE | re.IGNORECASE)
-    JOURNAL_NESTED_QUERY_PATH = 'publication_info'
     AUTHORS_NESTED_QUERY_PATH = 'authors'
+    DATE_NESTED_FIELDS = [
+        'publication_info.year',
+    ]
+    DATE_NESTED_QUERY_PATH = 'publication_info'
+    JOURNAL_NESTED_QUERY_PATH = 'publication_info'
     TITLE_SYMBOL_INDICATING_CHARACTER = ['-', '(', ')']
+    NESTED_FIELDS = ['authors', 'publication_info']
+    RECORD_RELATION_FIELD = 'related_records.relation'
+
     # ################
 
     # #### Helpers ####
@@ -184,60 +203,92 @@ class ElasticSearchVisitor(Visitor):
         """Generates a query handling specifically authors.
 
         Notes:
-            The match query is generic enough to return many results. Then, using the filter clause we truncate these
-            so that we imitate legacy's behaviour on returning more "exact" results. E.g. Searching for `Smith, John`
-            shouldn't return papers of 'Smith, Bob'.
+            There are three main cases:
 
-            Additionally, doing a ``match`` with ``"operator": "and"`` in order to be even more exact in our search, by
-            requiring that ``full_name`` field contains both
+            1) ``a Smith``
+            This will just generate a ``match`` query on ``last_name``
+
+            2) ``a John Smith``
+             This will just generate a ``match`` query on ``last_name`` and  a ``prefix`` query on ``first_name``
+             and a ``match`` query on the initial ``J``. This will return results from ``Smith, John`` and ``Smith, J``
+             but not from ``Smith, Jane``.
+
+            3) ``a J Smith``
+            This will just generate a ``match`` query on ``last_name`` and a match query on ``first_name.initials``.
+
+            Please note, cases such as ``J.D.`` have been properly handled by the tokenizer.
         """
-        name_variations = [name_variation.lower()
-                           for name_variation
-                           in generate_minimal_name_variations(author_name)]
+        parsed_name = ParsedName(author_name)
 
-        # When the query contains sufficient data, i.e. full names, e.g. ``Mele, Salvatore`` (and not ``Mele, S`` or
-        # ``Mele``) we can improve our filtering in order to filter out results containing records with authors that
-        # have the same non lastnames prefix, e.g. 'Mele, Samuele'.
-        if author_name_contains_fullnames(author_name):
-            specialized_author_filter = [
-                {
-                    'bool': {
-                        'must': [
-                            {
-                                'term': {ElasticSearchVisitor.AUTHORS_NAME_VARIATIONS_FIELD: names_variation[0]}
-                            },
-                            generate_match_query(
-                                ElasticSearchVisitor.KEYWORD_TO_ES_FIELDNAME['author'],
-                                names_variation[1],
-                                with_operator_and=True
-                            )
-                        ]
-                    }
-                } for names_variation
-                in product(name_variations, name_variations)
-            ]
-
-        else:
-            # In the case of initials or even single lastname search, filter with only the name variations.
-            specialized_author_filter = [
-                {'term': {ElasticSearchVisitor.AUTHORS_NAME_VARIATIONS_FIELD: name_variation}}
-                for name_variation in name_variations
-            ]
-
-        return {
-            'bool': {
-                'filter': {
-                    'bool': {
-                        'should': specialized_author_filter
-                    }
-                },
-                'must': {
-                    'match': {
-                        ElasticSearchVisitor.KEYWORD_TO_ES_FIELDNAME['author']: author_name
+        def _match_query_with_analyzer(field, value):
+            return {
+                "match": {
+                    ElasticSearchVisitor.KEYWORD_TO_ES_FIELDNAME[field]: {
+                        "query": value,
+                        "analyzer": "names_initials_analyzer"
                     }
                 }
             }
-        }
+
+        def _match_query_with_and_operator(field, value):
+            return {
+                'match': {
+                    ElasticSearchVisitor.KEYWORD_TO_ES_FIELDNAME[field]: {
+                        'query': value,
+                        'operator': 'AND'
+                    }
+                }
+            }
+
+        def _match_phrase_prefix_query(field, value):
+            return {
+                "match_phrase_prefix": {
+                    ElasticSearchVisitor.KEYWORD_TO_ES_FIELDNAME[field]: {
+                        "query": value,
+                        "analyzer": "names_analyzer"
+                    }
+                }
+            }
+
+        if len(parsed_name) == 1 and '.' not in parsed_name.first:
+            # ParsedName returns first name if there is only one name i.e. `Smith`
+            # in our case we consider it as a lastname
+            last_name = parsed_name.first
+            query = _match_query_with_and_operator("author_last_name", last_name)
+            return generate_nested_query(ElasticSearchVisitor.AUTHORS_NESTED_QUERY_PATH, query)
+
+        bool_query_build = []
+        bool_query_build.append(
+            _match_query_with_and_operator("author_last_name", parsed_name.last)
+        )
+
+        should_query = []
+        first_names = retokenize_first_names(parsed_name.first_list)
+        for name in first_names:
+            name_query = []
+            if is_initial_of_a_name(name):
+                name_query.append(
+                    _match_query_with_analyzer("author_first_name", name)
+                )
+            else:
+                name_query.extend([
+                    _match_phrase_prefix_query("author_first_name", name),
+                    _match_query_with_analyzer("author_first_name", name)
+                ])
+            should_query.append(
+                wrap_queries_in_bool_clauses_if_more_than_one(
+                    name_query, use_must_clause=False)
+            )
+
+        bool_query_build.append(
+            wrap_queries_in_bool_clauses_if_more_than_one(
+                should_query, use_must_clause=False)
+        )
+
+        query = wrap_queries_in_bool_clauses_if_more_than_one(
+            bool_query_build, use_must_clause=True
+        )
+        return generate_nested_query(ElasticSearchVisitor.AUTHORS_NESTED_QUERY_PATH, query)
 
     def _generate_exact_author_query(self, author_name_or_bai):
         """Generates a term query handling authors and BAIs.
@@ -251,9 +302,10 @@ class ElasticSearchVisitor(Visitor):
             'j. smith', 'J Smith', 'J. Smith'.
         """
         if ElasticSearchVisitor.BAI_REGEX.match(author_name_or_bai):
+            bai = author_name_or_bai.lower()
             query = self._generate_term_query(
-                '.'.join((ElasticSearchVisitor.AUTHORS_BAI_FIELD, FieldVariations.raw)),
-                author_name_or_bai
+                '.'.join((ElasticSearchVisitor.AUTHORS_BAI_FIELD, FieldVariations.search)),
+                bai
             )
         else:
             author_name = normalize('NFKC', normalize_name(author_name_or_bai)).lower()
@@ -334,7 +386,7 @@ class ElasticSearchVisitor(Visitor):
             See: https://github.com/inspirehep/inspire-query-parser/issues/79
             Otherwise, we query both ``document_type`` and ``publication_info``.
         """
-        mapping_for_value = self.TYPECODE_VALUE_TO_FIELD_AND_VALUE_PAIRS_MAPPING.get(value, None)
+        mapping_for_value = self.TYPECODE_VALUE_TO_FIELD_AND_VALUE_PAIRS_MAPPING.get(value.lower(), None)
 
         if mapping_for_value:
             return generate_match_query(*mapping_for_value, with_operator_and=True)
@@ -368,11 +420,21 @@ class ElasticSearchVisitor(Visitor):
 
         return query
 
-    # TODO Move it to visitor utils
-    def _generate_term_query(self, fieldname, value):
+    # TODO Move it to visitor utils and write tests for it.
+    def _generate_term_query(self, fieldname, value, boost=None):
+        if not boost:
+            return {
+                'term': {
+                    fieldname: value
+                }
+            }
+
         return {
             'term': {
-                fieldname: value
+                fieldname: {
+                    'value': value,
+                    'boost': boost
+                }
             }
         }
 
@@ -414,11 +476,17 @@ class ElasticSearchVisitor(Visitor):
                 if not updated_operator_value_pairs:
                     break  # Malformed date
                 else:
-                    range_queries.append({
+                    range_query = {
                         'range': {
                             fieldname: updated_operator_value_pairs
                         }
-                    })
+                    }
+
+                    range_queries.append(
+                        generate_nested_query(ElasticSearchVisitor.DATE_NESTED_QUERY_PATH, range_query)
+                        if fieldname in ElasticSearchVisitor.DATE_NESTED_FIELDS
+                        else range_query
+                    )
         else:
             range_queries = [{
                     'range': {
@@ -445,8 +513,8 @@ class ElasticSearchVisitor(Visitor):
             query_str = ' '.join([word.strip(':') for word in data.children])
 
         return {
-            'query_string': {
-                'default_field': '_all',
+            'simple_query_string': {
+                'fields': ['_all'],
                 'query': query_str
             }
         }
@@ -608,7 +676,38 @@ class ElasticSearchVisitor(Visitor):
         return self._generate_range_queries(force_list(fieldnames), {'lte': node.op.value})
 
     def visit_nested_keyword_op(self, node):  # TODO Cannot be completed as of yet.
-        raise NotImplementedError('Nested keyword queries aren\'t implemented yet.')
+        # FIXME: quick and dirty implementation of refersto:recid:<recid>
+        if node.left.value == 'refersto':
+            right = node.right
+            if hasattr(right, 'left') and hasattr(right, 'right') and right.left.value == 'control_number':
+                recid = right.right.value
+                citing_records_query = generate_match_query(
+                    ElasticSearchVisitor.KEYWORD_TO_ES_FIELDNAME['refersto'],
+                    recid,
+                    with_operator_and=False
+                )
+                records_with_collection_literature_query = generate_match_query(
+                    '_collections',
+                    'Literature',
+                    with_operator_and=False
+                )
+                superseded_records_query = generate_match_query(
+                    ElasticSearchVisitor.RECORD_RELATION_FIELD,
+                    'successor',
+                    with_operator_and=False
+                )
+                self_citation = generate_match_query(
+                    "control_number",
+                    recid,
+                    with_operator_and=False
+                )
+                return {
+                    'bool': {
+                        'must': [citing_records_query, records_with_collection_literature_query],
+                        'must_not': [superseded_records_query, self_citation]
+                    }
+                }
+        logger.warning('Nested keyword queries aren\'t implemented yet, except refersto:recid:<recid>')
 
     def visit_keyword(self, node):
         # If no keyword is found, return the original node value (case of an unknown keyword).
@@ -628,9 +727,15 @@ class ElasticSearchVisitor(Visitor):
                 query_bai_field_if_dots_in_name=True
             )
 
-            return self._generate_query_string_query(node.value,
-                                                     fieldnames=bai_fieldnames or fieldnames,
-                                                     analyze_wildcard=True)
+            query = self._generate_query_string_query(
+                node.value,
+                fieldnames=bai_fieldnames or fieldnames,
+                analyze_wildcard=True
+            )
+
+            if ElasticSearchVisitor.KEYWORD_TO_ES_FIELDNAME['author'] == fieldnames:
+                return generate_nested_query(ElasticSearchVisitor.AUTHORS_NESTED_QUERY_PATH, query)
+            return query
         else:
             if isinstance(fieldnames, list):
                 if ElasticSearchVisitor.KEYWORD_TO_ES_FIELDNAME['date'] == fieldnames:
@@ -656,7 +761,8 @@ class ElasticSearchVisitor(Visitor):
                     )
                     if bai_fieldnames:
                         if len(bai_fieldnames) == 1:
-                            return {"match": {bai_fieldnames[0]: node.value}}
+                            query = {"match": {bai_fieldnames[0]: node.value}}
+                            return generate_nested_query(ElasticSearchVisitor.AUTHORS_NESTED_QUERY_PATH, query)
                         else:
                             # Not an exact BAI pattern match, but node's value looks like BAI (no spaces and dots),
                             # e.g. `S.Mele`. In this case generate a partial match query.
@@ -676,6 +782,22 @@ class ElasticSearchVisitor(Visitor):
                 elif ElasticSearchVisitor.KEYWORD_TO_ES_FIELDNAME['type-code'] == fieldnames:
                     return self._generate_type_code_query(node.value)
 
+                elif ElasticSearchVisitor.KEYWORD_TO_ES_FIELDNAME['affiliation'] == fieldnames:
+                    query = generate_match_query(
+                        ElasticSearchVisitor.KEYWORD_TO_ES_FIELDNAME['affiliation'],
+                        node.value,
+                        with_operator_and=True
+                    )
+                    return generate_nested_query(ElasticSearchVisitor.AUTHORS_NESTED_QUERY_PATH, query)
+
+                elif fieldnames not in ElasticSearchVisitor.KEYWORD_TO_ES_FIELDNAME.values():
+                    colon_value = ':'.join([fieldnames, node.value])
+                    given_field_query = generate_match_query(fieldnames, node.value, with_operator_and=True)
+                    texkey_query = self._generate_term_query('texkeys.raw', colon_value, boost=2.0)
+                    _all_field_query = generate_match_query('_all', colon_value, with_operator_and=True)
+                    query = wrap_queries_in_bool_clauses_if_more_than_one([given_field_query, texkey_query, _all_field_query], use_must_clause=False)
+                    return wrap_query_in_nested_if_field_is_nested(query, fieldnames, ElasticSearchVisitor.NESTED_FIELDS)
+
                 return generate_match_query(fieldnames, node.value, with_operator_and=True)
 
     def visit_exact_match_value(self, node, fieldnames=None):
@@ -701,9 +823,21 @@ class ElasticSearchVisitor(Visitor):
         )
 
         if ElasticSearchVisitor.KEYWORD_TO_ES_FIELDNAME['date'] == fieldnames:
-            term_queries = [{'term': {field: _truncate_date_value_according_on_date_field(field, node.value).dumps()}}
-                            for field
-                            in fieldnames]
+            term_queries = []
+            for field in fieldnames:
+                term_query =  \
+                    {'term': {field: _truncate_date_value_according_on_date_field(field, node.value).dumps()}}
+
+                term_queries.append(
+                    generate_nested_query(ElasticSearchVisitor.DATE_NESTED_QUERY_PATH, term_query)
+                    if field in ElasticSearchVisitor.DATE_NESTED_FIELDS
+                    else term_query
+                )
+        elif ElasticSearchVisitor.KEYWORD_TO_ES_FIELDNAME['author'] in fieldnames:
+            term_queries = [
+                generate_nested_query(ElasticSearchVisitor.AUTHORS_NESTED_QUERY_PATH, {'term': {field: node.value}})
+                for field in (bai_fieldnames or fieldnames)
+            ]
         else:
             term_queries = [{'term': {field: node.value}} for field in (bai_fieldnames or fieldnames)]
 
@@ -740,13 +874,23 @@ class ElasticSearchVisitor(Visitor):
             query_bai_field_if_dots_in_name=True
         )
 
-        return self._generate_query_string_query(value,
-                                                 fieldnames=bai_fieldnames or fieldnames,
-                                                 analyze_wildcard=True)
+        query = self._generate_query_string_query(value,
+                                                  fieldnames=bai_fieldnames or fieldnames,
+                                                  analyze_wildcard=True)
+        if (bai_fieldnames and ElasticSearchVisitor.KEYWORD_TO_ES_FIELDNAME['author'] in bai_fieldnames) \
+                or (fieldnames and ElasticSearchVisitor.KEYWORD_TO_ES_FIELDNAME['author'] in fieldnames):
+            return generate_nested_query(ElasticSearchVisitor.AUTHORS_NESTED_QUERY_PATH, query)
+
+        return query
 
     def visit_regex_value(self, node, fieldname):
-        return {
+        query = {
             'regexp': {
                 fieldname: node.value
             }
         }
+
+        if ElasticSearchVisitor.KEYWORD_TO_ES_FIELDNAME['author'] == fieldname:
+            return generate_nested_query(ElasticSearchVisitor.AUTHORS_NESTED_QUERY_PATH, query)
+
+        return query
